


function fP!(fp, dws::DiracWorkspace, prop::AbstractArray,lp::SpaceParm)

    function krnl_fP!(sum,pro)
            b=Int64(CUDA.threadIdx().x)
            r=Int64(CUDA.blockIdx().x)    
            I = point_coord((b,r), lp)

            sum[I] = norm2(pro[b,r])/2
        return nothing
    end

    CUDA.@sync begin
        CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_fP!(dws.cs,prop)
    end

    fp += reshape(Array(CUDA.reduce(+, dws.cs; dims=(1,2,3))),lp.iL[end])

    return nothing
end

function fA!(fa, dws::DiracWorkspace, prop::AbstractArray,lp::SpaceParm)

    function krnl_fA!(sum,pro)
            b=Int64(CUDA.threadIdx().x)
            r=Int64(CUDA.blockIdx().x)    
            I = point_coord((b,r), lp)

            sum[I] = -dot(pro[b,r],dmul(Gamma{4},pro[b,r]))/2
        return nothing
    end
    
    CUDA.@sync begin
        CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_fA!(dws.cs,prop)
    end
    fa += reshape(Array(CUDA.reduce(+, dws.cs; dims=(1,2,3))),lp.iL[end])
    return nothing
end

function kV!(kv, dws::DiracWorkspace, prop::AbstractArray,lp::SpaceParm)

    function krnl_kV!(sum,pro)
            b=Int64(CUDA.threadIdx().x)
            r=Int64(CUDA.blockIdx().x)    
            I = point_coord((b,r), lp)

            sum[I] = -(norm2(dmul(Gamma{6},pro[b,r])) + norm2(dmul(Gamma{7},pro[b,r])) + norm2(dmul(Gamma{8},pro[b,r])))/6
        return nothing
    end
    
    CUDA.@sync begin
        CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_kV!(dws.cs,prop)
    end
    kv += reshape(Array(CUDA.reduce(+, dws.cs; dims=(1,2,3))),lp.iL[end])
    return nothing
end

function lV!(lv, dws::DiracWorkspace, prop::AbstractArray,lp::SpaceParm)

    function krnl_lV!(sum,pro)
            b=Int64(CUDA.threadIdx().x)
            r=Int64(CUDA.blockIdx().x)    
            I = point_coord((b,r), lp)

            sum[I] = im*(dot(dmul(Gamma{6},pro[b,r]),dmul(Gamma{5},dmul(Gamma{10},pro[b,r])))
                       + dot(dmul(Gamma{7},pro[b,r]),dmul(Gamma{5},dmul(Gamma{11},pro[b,r])))
                       + dot(dmul(Gamma{8},pro[b,r]),dmul(Gamma{5},dmul(Gamma{12},pro[b,r]))))/6
        return nothing
    end
    
    CUDA.@sync begin
        CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_lV!(dws.cs,prop)
    end
    lv += reshape(Array(CUDA.reduce(+, dws.cs; dims=(1,2,3))),lp.iL[end])
    return nothing
end
